/*
** main.c
**
** Nokia 3310 Example Program
** Target: ATMEGA :: AVR-GCC
**
** Written by Tony Myatt - 2007
** Quantum Torque - www.quantumtorque.com
*/
#define F_CPU 12000000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BOARD_START_X 4
#define BOARD_START_Y 11

#define CELL_SIZE 4

#define BOARD_WIDTH	22
#define BOARD_HEIGHT 14

/* Array cu starea board-ului de joc */
static unsigned char game_board[BOARD_WIDTH][BOARD_HEIGHT];

#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3
#define NONE 4
#define FOOD 5

typedef struct coord {
	unsigned char x, y;
} coord_t;

static coord_t head, tail;

/* Coordonate relative pentru deplasare */
static char dx[4] = {0, 0, -1, +1};
static char dy[4] = {-1, +1, 0, 0};

static int score = 0;


/* Deseneaza in screenbuffer o celula din game_board */
void draw_board_cell(unsigned char i, unsigned char j)
{
	char x, y, k, l;

	x = BOARD_START_X + i * CELL_SIZE;
	y = BOARD_START_Y + j * CELL_SIZE;

	/* Desenez ceilalti pixelii */
	switch (game_board[i][j]) {

		case UP: case DOWN:
			for (k = y; k < y + CELL_SIZE; k++) {
				lcd_set_pixel(x + 1, k);
				lcd_set_pixel(x + 2, k);
			}
			break;

		case LEFT : case RIGHT:
			for (k = x; k < x + CELL_SIZE; k++) {
				lcd_set_pixel(k, y + 1);
				lcd_set_pixel(k, y + 2);
			}
			break;

		case NONE:
			for (k = x; k < x + CELL_SIZE; k++)
				for (l = y; l < y + CELL_SIZE; l++)
					lcd_clear_pixel(k, l);
			break;

		case FOOD:
			for (k = x + 1; k < x + 3; k++) {
				for (l = y + 1; l < y + 3; l++) {
					lcd_set_pixel(k, l);
				}
			}
			break;
	}
}

/* Deseneaza o celula in care s-a luat o curba */
void draw_turning_cell(unsigned char i, unsigned char j, unsigned char tip)
{
	char x, y, k, l;

	x = BOARD_START_X + i * CELL_SIZE;
	y = BOARD_START_Y + j * CELL_SIZE;

	/* Mai intai sterg celula */
	for (k = x; k < x + CELL_SIZE; k++)
		for (l = y; l < y + CELL_SIZE; l++)
			lcd_clear_pixel(k, l);

	/* Setez centrul */
	for (k = x + 1; k < x + 3; k++)
		for (l = y + 1; l < y + 3; l++)
			lcd_set_pixel(k, l);

	switch (tip) {
		case 0:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x, y + k);
				lcd_set_pixel(x + k, y + 3);
			}
			break;

		case 1:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x + 3, y + k);
				lcd_set_pixel(x + k, y + 3);
			}
			break;

		case 2:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x, y + k);
				lcd_set_pixel(x + k, y);
			}
			break;

		case 3:
			for (k = 1; k < 3; k++) {
				lcd_set_pixel(x + k, y);
				lcd_set_pixel(x + 3, y + k);
			}
			break;
	}

}


void draw_game_board()
{
	for (unsigned char i = 0; i < BOARD_WIDTH; i++) {
		for (unsigned char j = 0; j < BOARD_HEIGHT; j++) {
			draw_board_cell(i, j);
		}
	}
}

/* Plaseaza un FOOD aleator */
void generate_food()
{
	unsigned char x, y;
	do {
		x = rand() % BOARD_WIDTH;
		y = rand() % BOARD_HEIGHT;
	} while (game_board[x][y] != NONE);

	game_board[x][y] = FOOD;
	draw_board_cell(x, y);
}

/* Initializeaza jocul */
void init_game_board()
{
	unsigned char i, j;
	for (i = 0; i < BOARD_WIDTH; i++)
		for (j = 0; j < BOARD_HEIGHT; j++)
			game_board[i][j] = NONE;

	// Initializare head si tail
	head.x = BOARD_WIDTH / 2 + 2;
	tail.x = BOARD_WIDTH / 2 - 2;
	head.y = tail.y = BOARD_HEIGHT / 2;

	// Desenare sarpe
	for (i = tail.x; i <= head.x; i++) {
		game_board[i][head.y] = RIGHT;
	}

	generate_food();

}

void game_over()
{	
	lcd_clear();
	lcd_goto_xy(1, 3);
	printf("GAME OVER");
	lcd_goto_xy(1, 4);
	printf("Press RESET.");
	while (1);
}

/* Updateaza o valoare. Face si wrapping */
void update_value(unsigned char *v, char dv, unsigned char max)
{
	if (*v + dv == max) {
		*v = 0;
		return;
	} else if (*v == 0 && dv == -1) {
		*v = max - 1;
		return;
	}

	*v += dv;
}

/* Muta sarpele inainte */
void go_forward()
{
	unsigned char current_tail, current_head;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];
	
	game_board[tail.x][tail.y] = NONE;

	/* Avansez capul */
	update_value(&head.x, dx[current_head], BOARD_WIDTH);
	update_value(&head.y, dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 5;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			update_value(&tail.x, dx[current_tail], BOARD_WIDTH);
			update_value(&tail.y, dy[current_tail], BOARD_HEIGHT);
		} else { // Este o bucata din sarpe; GAME OVER.
			game_over();
		}
	}

	game_board[head.x][head.y] = current_head;

	draw_board_cell(head.x, head.y);
}

/* Turn helper */
static unsigned char turn_helper[2][4] = {{0, 3, 1, 2}, {1, 2, 3, 0}};
/* Coordonate relative pentru deplasarea la stanga */
static char left_dx[4] = {-1, 1, 0, 0};
static char left_dy[4] = {0, 0, 1, -1};
static char left_next[4] = {LEFT, RIGHT, DOWN, UP};

/* Muta sarpele la stanga */
void go_left()
{
	unsigned char current_head, current_tail;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];

	game_board[tail.x][tail.y] = NONE;
	game_board[head.x][head.y] = left_next[current_head];

	/* Avansez capul */
	draw_turning_cell(head.x, head.y, turn_helper[0][current_head]);
	update_value(&head.x, left_dx[current_head], BOARD_WIDTH);
	update_value(&head.y, left_dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 5;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			update_value(&tail.x, dx[current_tail], BOARD_WIDTH);
			update_value(&tail.y, dy[current_tail], BOARD_HEIGHT);
		} else { // Este o bucata din sarpe; GAME OVER.
			game_over();
		}
	}

	game_board[head.x][head.y] = left_next[current_head];

	draw_board_cell(head.x, head.y);
}

/* Coordonate relative pentru deplasare la dreapta */
static char right_dx[4] = {1, -1, 0, 0};
static char right_dy[4] = {0, 0, -1, +1};
static char right_next[4] = {RIGHT, LEFT, UP, DOWN};

/* Muta sarpele la stanga */
void go_right()
{
	unsigned char current_head, current_tail;

	current_head = game_board[head.x][head.y];
	current_tail = game_board[tail.x][tail.y];

	game_board[tail.x][tail.y] = NONE;
	game_board[head.x][head.y] = right_next[current_head];

	/* Avansez capul */
	draw_turning_cell(head.x, head.y, turn_helper[1][current_head]);
	update_value(&head.x, right_dx[current_head], BOARD_WIDTH);
	update_value(&head.y, right_dy[current_head], BOARD_HEIGHT);

	if (game_board[head.x][head.y] == FOOD) {
		game_board[tail.x][tail.y] = current_tail;
		score += 5;
		generate_food();
	} else {
		if (game_board[head.x][head.y] == NONE) {
			draw_board_cell(tail.x, tail.y);
			update_value(&tail.x, dx[current_tail], BOARD_WIDTH);
			update_value(&tail.y, dy[current_tail], BOARD_HEIGHT);
		} else { // Este o bucata din sarpe; GAME OVER.
			game_over();
		}
	}

	game_board[head.x][head.y] = right_next[current_head];

	draw_board_cell(head.x, head.y);
}



/* Setez cele 2 butoane ca input */
void input_init()
{
	DDRB &= ~ (1 << PB2);
	PORTB |= (1 << PB2);

	DDRD &= ~ (1 << PD1);
	PORTD |= (1 << PD1);
}

static char pressed_left = 0;
static char pressed_right = 0;

/* main -- Program starts here */
int main(void)
{	
	// Setup LCD

	srand(17);

	input_init();

	lcd_init();
	lcd_contrast(0x40);

	lcd_clear();

	init_game_board();
	draw_game_board();

	lcd_draw_border();

	lcd_draw_buffer();

	while (1) {
		if (pressed_left) {
			go_left();
			pressed_left = 0;
		} else if (pressed_right) {
			go_right();
			pressed_right = 0;
		} else {
			go_forward();
		}

		lcd_draw_border();
		lcd_draw_buffer();
		lcd_draw_score(score);

		for (int i = 0; i < 10000; i++) {
			for (int j = 0; j < 70; j++) {
				if (!(PIND & (1 << PD1))) {
					pressed_right = 1;
				}

				if (!(PINB & (1 << PB2))) {
					pressed_left = 1;
				}
			}
		}
	}
}
